﻿/**
ログ。

親から子へログオブジェクトが渡されていくので擬似的な階層出力になる。
基本的には以下の形でなんらかに出力。
$(PREFORMATTED
YYYY/MM/DDThh:mm:ss.ms module(line) [Name:Number] Message
YYYY/MM/DDThh:mm:ss.ms module(line) [Name:Number>Name:Number] Message
YYYY/MM/DDThh:mm:ss.ms module(line) [Name:Number>Name:Number>Name:Number] Message
)

History:
	$(HIST_VER 1.100,
		$(HIST_NEW)
	)
*/
module etc.negui.system.utility.log;

debug import std.stdio: wl = writefln;
debug(log) void main() {}

import std.stream;

import etc.negui.system.base;

///
class LogException: NeGuiException {
	mixin MixInNeGuiException;
}

/**
*/
struct LOGGERINFO {
	size_t number;
	Text   name;

	const ref Text toText() {
		if(name.length) {
			return Text("%s:%08s", name.data, number);
		} else {
			return Text("%08s", number);
		}
	}

	private this(const ref LOGGERINFO LoggerInfo) {
		number = LoggerInfo.number;
		name   = LoggerInfo.name.dup;
	}
}

/**
ログの継承用。
*/
interface ILogger {
	/**
	write時の挙動。
	*/
	protected void DoWrite(in Text Message);
	/**
	子供の作成。
	
	scopeが良かった。
	*/
	ILogger logger(in Text Name, bool Inheritance=true);

	/**
	書き込み

	基本的にはsuper.write呼び出せば問題ないかと。
	*/
	const(Text) write(lazy const(Text) Message, string _file=__FILE__, int _line=__LINE__);
}

/**
ログ。
*/
abstract class Log: ILogger {
	protected {
		LOGGERINFO* Owner;
		LOGGERINFO Me;
	}
	/**
	*/
	this(in Text Name, const LOGGERINFO* OwnerLogger=null) {
		if(OwnerLogger) {
			Owner = new LOGGERINFO(*OwnerLogger);
		}
		Me.name = Name.dup;
	}
	///
	bool enable;

	private {
		const ref Text Format(in string FilePath, in size_t LineNumber, in Text Message) {
			auto time=new DateTime();
			time.mask |= DateTime.MASK.MS;
			time.format = Text("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[MS]");

			return Text(
				"%s %s(%s) [%s] %s",
				time,
				FilePath,
				LineNumber,
				Information,
				Message
			);
		}
		const ref Text OwnerText()
		in {
			assert(Owner);
		}
		body {
			return Text.wrap(Owner.toText.data ~ '/');
		}

		const Text Information() {
			if(Owner) {
				return OwnerText ~ Me.toText;
			} else {
				return Me.toText;
			}
		}
	}

	/***/
	override const(Text) write(lazy const(Text) Message, string _file=__FILE__, int _line=__LINE__)
	out(r) {
		if(!enable) {
			assert(!r.length, r.str);
		}
	}
	body {
		if(enable) {
			Me.number++;
			const FormatedMessage=Format(_file, _line, Message);
			DoWrite(FormatedMessage);

			return FormatedMessage;
		} else {
			return Text.init;
		}
	}

	/**
	logger用LOGGERINFO情報設定。
	*/
	protected const LOGGERINFO GetMeToOwner() {
		LOGGERINFO LoggerInfo;

		LoggerInfo.name   = OwnerText ~ Me.name;
		LoggerInfo.number = Me.number;

		return LoggerInfo;
	}
}

/**
ファイル出力ログ。

History:
	$(HIST_VER 1.120,
		$(HIST_F
			コンストラクタ時にファイルを作らないように変更。
			そのかわりSMixInLogを使用する前には最上位のLoggerで出力を行わなきゃ死ぬ。
		)
	)
*/
class Logger: Log, ILogger {
	protected {
		EndianStream stream;
		Text LogFilePath;
	}
	private this(EndianStream stream, in Text Name, const ref LOGGERINFO OwnerLogger) {
		this.stream = stream;
		super(Name, &OwnerLogger);
	}
	/***/
	this(in Text LogFilePath, in Text Name) {
		/+
		try {
			stream = new EndianStream(new BufferedFile(LogFilePath.str, FileMode.OutNew));
		} catch(Exception e) {
			throw new LogException(Text("LogFilePath -> %s", LogFilePath), e);
		}
		+/
		super(Name);
		
		this.LogFilePath = LogFilePath.dup;
	}
	/***/
	void delegate(in Text Message) exWrite;

	/***/
	final void close() {
		stream.close;
		delete stream;
	}

	~this() {
		if(!Owner && stream) {
			close;
		}
	}

	override {
		/**
		SMixInLog使うときは注意。
		*/
		protected void DoWrite(in Text Message) {
			if(!stream) {
				try {
					stream = new EndianStream(new BufferedFile(LogFilePath.str, FileMode.OutNew));
				} catch(Exception e) {
					throw new LogException(Text("LogFilePath -> %s", LogFilePath), e);
				}
			}
			
			stream.writeLineW(Message.data);
			stream.flush();
		}
		/***/
		Logger logger(in Text Name, bool Inheritance=true) {
			LOGGERINFO LoggerInfo;

			if(Owner) {
				LoggerInfo = GetMeToOwner;
			} else {
				LoggerInfo = Me;
			}

			auto log=new Logger(stream, Name, LoggerInfo);
			//log.LogFilePath = LogFilePath;
			if(Inheritance) {
				log.exWrite = exWrite;
				log.enable  = enable;;
			}

			return log;
		}
		/***/
		const(Text) write(lazy const(Text) Message, string _file=__FILE__, int _line=__LINE__) {
			const FormatedMessage=super.write(Message, _file, _line);
			if(enable && exWrite) {
				exWrite(FormatedMessage);
			}
			return FormatedMessage;
		}
	}
}

debug(log) unittest {
	auto l=new Logger(Text("log.log"), Text("asd"));
	l.exWrite = (in Text Message) {
		wl("%s", Message);
	};
	l.enable =true;
	l.write(Text(123));
	auto l2=l.logger(Text("ASD"));
	l2.write(Text(456));
	auto l3=l2.logger(Text("xyz"));
	l3.write(Text(789));
	l2.write(Text(11111111111));
	l3.write(Text(789));
	l3.write(Text(789));
	l3.write(Text(789));
	l.write(Text(123));
	l2.write(Text(456));
	l.write(Text(123));
	l.write(Text(123));
	l3.write(Text(789));
	delete l;
}
/***/
string SMixInLog(string LogTitle, string OwnerLogName="OwnerLog", string ScopeLogName="log") {
	return `
	
	scope `~ ScopeLogName ~`=` ~ OwnerLogName ~ `.logger(Text("` ~ LogTitle ~ `"), true);
	`~ ScopeLogName ~`.write(Text("start..."));
	scope(exit)    `~ ScopeLogName ~`.write(Text("end."));
	scope(success) `~ ScopeLogName ~`.write(Text("success."));
	scope(failure) `~ ScopeLogName ~`.write(Text("failure."));
	
	`;
}
debug(log) unittest {
	scope l=new Logger(Text("_log.log"), Text(""));
	l.enable =true;
	l.write(Text.init);
	{
		mixin(SMixInLog("BunBun", `l`));
		log.write(Text(`(n‘∀‘)η ﾔｧｰｯﾎｫｰ`));
		throw new LogException(Text(__VENDOR__));
	}
}
